# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import Blueprint, Flask, Response, request

from rucio.common.constants import DEFAULT_VO, HTTPMethod
from rucio.common.exception import AccessDenied, DataIdentifierNotFound, OpenDataDataIdentifierAlreadyExists, OpenDataDataIdentifierNotFound
from rucio.common.utils import render_json
from rucio.core.opendata import validate_opendata_did_state
from rucio.gateway import opendata
from rucio.web.rest.flaskapi.authenticated_bp import AuthenticatedBlueprint
from rucio.web.rest.flaskapi.v1.common import ErrorHandlingMethodView, check_accept_header_wrapper_flask, generate_http_error_flask, json_parameters, param_get, parse_scope_name, response_headers


class OpenDataView(ErrorHandlingMethodView):
    @staticmethod
    def get_helper(public: bool) -> "Response":
        """
        Helper function to list Open Data DIDs. To be used by both authenticated and unauthenticated views.
        """
        try:
            state = request.args.get("state", default=None) if not public else "PUBLIC"
            if state is not None:
                state = validate_opendata_did_state(state)

            limit = request.args.get("limit", default=None, type=int)  # type: ignore
            offset = request.args.get("offset", default=None, type=int)  # type: ignore
            result = opendata.list_opendata_dids(limit=limit, offset=offset, state=state)
            result = render_json(**result)
            return Response(result, status=200, mimetype='application/json')
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except Exception as error:
            return generate_http_error_flask(400, error)

    @check_accept_header_wrapper_flask(["application/json"])
    def get(self) -> "Response":
        """
        ---
        summary: List Opendata DIDs
        description: "Retrieves a list of Opendata Data Identifiers (DIDs). Supports optional query parameters for pagination and filtering by state."
        tags:
          - Opendata
        parameters:
          - name: limit
            in: query
            description: "Maximum number of results to return."
            schema:
              type: integer
            required: false
            style: form
          - name: offset
            in: query
            description: "Number of items to skip before starting to collect the result set."
            schema:
              type: integer
            required: false
            style: form
          - name: state
            in: query
            description: "Filter DIDs by their state (e.g., 'PUBLIC')."
            schema:
              type: string
            required: false
            style: form
        responses:
          200:
            description: "Successful retrieval of the list of Opendata DIDs."
            content:
              application/json:
                schema:
                  type: object
          401:
            description: "Access denied: Invalid authentication."
          400:
            description: "Invalid request or query parameters."
        """
        return self.get_helper(public=False)


class OpenDataDIDsView(ErrorHandlingMethodView):

    @staticmethod
    def get_helper(*, scope: str, name: str, public: bool) -> "Response":
        """
        Helper function to get Open Data DID information. To be used by both authenticated and unauthenticated views.
        """
        try:
            vo = request.environ.get("vo", DEFAULT_VO) if not public else DEFAULT_VO
            state = request.args.get("state", default=None) if not public else "public"
            if state is not None:
                state = validate_opendata_did_state(state)

            scope, name = parse_scope_name(f"{scope}/{name}", vo=vo)
            include_files = request.args.get("files", default="0").lower() == "1"
            include_metadata = request.args.get("meta", default="0").lower() == "1"
            include_doi = request.args.get("doi", default="1").lower() == "1"
            result = opendata.get_opendata_did(scope=scope, name=name, vo=vo,
                                               state=state,
                                               include_files=include_files,
                                               include_metadata=include_metadata,
                                               include_doi=include_doi,
                                               )

            result = render_json(**result)
            return Response(result, status=200, mimetype='application/json')
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except OpenDataDataIdentifierNotFound as error:
            return generate_http_error_flask(404, error)
        except Exception as error:
            return generate_http_error_flask(400, error)

    @check_accept_header_wrapper_flask(["application/json"])
    def get(self, scope: str, name: str) -> "Response":
        """
        ---
        summary: Get Opendata DID Information
        description: "Retrieves detailed Opendata information for the given scope and name. Supports optional query parameters to control the inclusion of files, metadata, and DOI information."
        tags:
          - Opendata
        parameters:
          - name: scope
            in: path
            description: "The scope of the data identifier."
            schema:
              type: string
            required: true
            style: simple
          - name: name
            in: path
            description: "The name of the data identifier."
            schema:
              type: string
            required: true
            style: simple
          - name: files
            in: query
            description: "Whether to include the list of files. '1' to include, '0' to exclude. Default is '0'."
            schema:
              type: string
              enum: ['0', '1']
            required: false
            style: form
          - name: meta
            in: query
            description: "Whether to include metadata. '1' to include, '0' to exclude. Default is '0'."
            schema:
              type: string
              enum: ['0', '1']
            required: false
            style: form
          - name: doi
            in: query
            description: "Whether to include the Digital Object Identifier (DOI). '1' to include, '0' to exclude. Default is '1'."
            schema:
              type: string
              enum: ['0', '1']
            required: false
            style: form
          - name: state
            in: query
            description: "Optional state filter for the data identifier."
            schema:
              type: string
            required: false
            style: form
        responses:
          200:
            description: "Successful retrieval of Opendata DID information."
            content:
              application/json:
                schema:
                  type: object
          401:
            description: "Access denied: Invalid authentication."
          404:
            description: "Data Identifier not found."
          400:
            description: "Invalid request or input parameters."
        """
        return self.get_helper(scope=scope, name=name, public=False)

    def post(self, scope: str, name: str) -> "Response":
        """
        ---
        summary: Register Opendata DID
        description: "Registers an existing DID as Opendata."
        tags:
          - Opendata
        parameters:
          - name: scope
            in: path
            description: "The scope of the data identifier to be registered."
            schema:
              type: string
            required: true
            style: simple
          - name: name
            in: path
            description: "The name of the data identifier to be registered."
            schema:
              type: string
            required: true
            style: simple
        responses:
          201:
            description: "Opendata DID successfully registered."
            content:
              application/json:
                schema:
                  type: string
                  enum: []
          400:
            description: "Invalid input: The provided scope/name is not valid."
          401:
            description: "Access denied: Invalid authentication."
          404:
            description: "Data Identifier not found."
          409:
            description: "Data Identifier already exists in the Opendata catalog."
        """
        vo = request.environ.get("vo", DEFAULT_VO)
        try:
            scope, name = parse_scope_name(f"{scope}/{name}", vo=vo)
            opendata.add_opendata_did(scope=scope, name=name, vo=vo)
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except DataIdentifierNotFound as error:
            return generate_http_error_flask(404, error)
        except OpenDataDataIdentifierAlreadyExists as error:
            return generate_http_error_flask(409, error)
        except Exception as error:
            return generate_http_error_flask(400, error)

        return Response(status=201, mimetype='application/json')

    def put(self, scope: str, name: str) -> "Response":
        """
        ---
        summary: Update Opendata DID
        description: "Updates the properties of an existing Opendata DID."
        tags:
          - Opendata
        parameters:
          - name: scope
            in: path
            description: "The scope of the data identifier to be updated."
            schema:
              type: string
            required: true
            style: simple
          - name: name
            in: path
            description: "The name of the data identifier to be updated."
            schema:
              type: string
            required: true
            style: simple
          - name: body
            in: body
            description: "JSON object containing the fields to update: 'state', 'meta', and/or 'doi'."
            required: false
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    state:
                      type: string
                      description: "New state for the DID."
                      enum: ["draft", "public", "suspended"]
                    meta:
                      type: object
                      description: "New metadata dictionary for the DID. Supports arbitrary JSON objects."
                      example: {"key": "value", "another_key": "another_value"}
                    doi:
                      type: string
                      description: "Digital Object Identifier (DOI) for the DID.
                      example: '10.1234/abcd.efgh'."
        responses:
          200:
            description: "Opendata DID successfully updated."
            content:
              application/json:
                schema:
                  type: string
                  enum: []
          400:
            description: "Invalid input or update parameters."
          401:
            description: "Access denied: Invalid authentication."
          404:
            description: "Data Identifier not found."
        """
        try:
            scope, name = parse_scope_name(f"{scope}/{name}", request.environ.get("vo", DEFAULT_VO))
            parameters = json_parameters()
            state = param_get(parameters, 'state', default=None)
            meta = param_get(parameters, 'meta', default=None)
            doi = param_get(parameters, 'doi', default=None)
            result = opendata.update_opendata_did(scope=scope,
                                         name=name,
                                         state=state,
                                         meta=meta,
                                         doi=doi,
                                         vo=request.environ.get("vo", DEFAULT_VO),
                                         )
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except OpenDataDataIdentifierNotFound as error:
            return generate_http_error_flask(404, error)
        except Exception as error:
            return generate_http_error_flask(400, error)

        return Response(status=200, mimetype='application/json', response=render_json(**result))

    def delete(self, scope: str, name: str) -> "Response":
        """
        ---
        summary: Delete Opendata DID
        description: "Deletes an entry in the Opendata catalog."
        tags:
          - Opendata
        parameters:
          - name: scope
            in: path
            description: "The scope of the data identifier to be deleted."
            schema:
              type: string
            required: true
            style: simple
          - name: name
            in: path
            description: "The name of the data identifier to be deleted."
            schema:
              type: string
            required: true
            style: simple
        responses:
          204:
            description: "Opendata DID successfully deleted. No content is returned."
          400:
            description: "Invalid input: The provided scope/name is not valid."
          401:
            description: "Access denied: Invalid authentication."
          404:
            description: "Data Identifier not found."
        """
        try:
            scope, name = parse_scope_name(f"{scope}/{name}", request.environ.get("vo", DEFAULT_VO))
            opendata.delete_opendata_did(scope=scope, name=name, vo=request.environ.get("vo", DEFAULT_VO))
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except OpenDataDataIdentifierNotFound as error:
            return generate_http_error_flask(404, error)
        except Exception as error:
            return generate_http_error_flask(400, error)
        return Response(status=204, mimetype='application/json')


def blueprint() -> "Blueprint":
    bp = AuthenticatedBlueprint("opendata", __name__, url_prefix="/opendata")

    opendata_view = OpenDataView.as_view("opendata")
    bp.add_url_rule("/dids", view_func=opendata_view, methods=[HTTPMethod.GET.value])

    opendata_did_view = OpenDataDIDsView.as_view("opendata_did")
    bp.add_url_rule("/dids/<scope>/<name>", view_func=opendata_did_view,
                    methods=[HTTPMethod.GET.value, HTTPMethod.POST.value, HTTPMethod.PUT.value, HTTPMethod.DELETE.value])

    bp.after_request(response_headers)

    return bp


def make_doc() -> "Flask":
    """ Only used for sphinx documentation """
    doc_app = Flask(__name__)
    doc_app.register_blueprint(blueprint())
    return doc_app




'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '38.4.0',
    'branch_nick': 'release-38-LTS',
    'revision_id': '945ab71be90243fe96148bb3bd13c1c3ae410765',
    'revno': 14030
}



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from json import dumps
from typing import TYPE_CHECKING, Any, Literal, Optional
from urllib.parse import quote_plus

from requests.status_codes import codes

from rucio.client.baseclient import BaseClient, choice
from rucio.common.constants import HTTPMethod, TransferLimitDirection
from rucio.common.utils import build_url

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence


class RequestClient(BaseClient):

    REQUEST_BASEURL = 'requests'

    def list_requests(
            self,
            src_rse: str,
            dst_rse: str,
            request_states: 'Sequence[str]'
    ) -> 'Iterator[dict[str, Any]]':
        """Return latest request details

        Returns
        -------
        request information
        """
        path = '/'.join([self.REQUEST_BASEURL, 'list']) + '?' + '&'.join(['src_rse={}'.format(src_rse), 'dst_rse={}'.format(
            dst_rse), 'request_states={}'.format(request_states)])
        url = build_url(choice(self.list_hosts), path=path)
        r = self._send_request(url, method=HTTPMethod.GET)

        if r.status_code == codes.ok:
            return self._load_json_data(r)
        else:
            exc_cls, exc_msg = self._get_exception(headers=r.headers, status_code=r.status_code, data=r.content)
            raise exc_cls(exc_msg)

    def list_requests_history(
            self,
            src_rse: str,
            dst_rse: str,
            request_states: 'Sequence[str]',
            offset: int = 0,
            limit: int = 100
    ) -> 'Iterator[dict[str, Any]]':
        """Return historical request details

        Returns
        -------
        request information
        """
        path = '/'.join([self.REQUEST_BASEURL, 'history', 'list']) + '?' + '&'.join(['src_rse={}'.format(src_rse), 'dst_rse={}'.format(
            dst_rse), 'request_states={}'.format(request_states), 'offset={}'.format(offset), 'limit={}'.format(limit)])
        url = build_url(choice(self.list_hosts), path=path)
        r = self._send_request(url, method=HTTPMethod.GET)

        if r.status_code == codes.ok:
            return self._load_json_data(r)
        else:
            exc_cls, exc_msg = self._get_exception(headers=r.headers, status_code=r.status_code, data=r.content)
            raise exc_cls(exc_msg)

    def list_request_by_did(
            self,
            name: str,
            rse: str,
            scope: Optional[str] = None
    ) -> 'Iterator[dict[str, Any]]':
        """Return latest request details for a DID
        Parameters
        ----------
        name:
            DID
        rse:
            Destination RSE name
        scope:
            rucio scope, defaults to None

        Raises
        -------
        exc_cls: from BaseClient._get_exception

        Returns
        -------
        request information
        """

        if scope is not None:
            path = '/'.join([self.REQUEST_BASEURL, quote_plus(scope), quote_plus(name), rse])
        url = build_url(choice(self.list_hosts), path=path)
        r = self._send_request(url, method=HTTPMethod.GET)

        if r.status_code == codes.ok:
            return next(self._load_json_data(r))
        else:
            exc_cls, exc_msg = self._get_exception(headers=r.headers, status_code=r.status_code, data=r.content)
            raise exc_cls(exc_msg)

    def list_request_history_by_did(
            self,
            name: str,
            rse: str,
            scope: Optional[str] = None
    ) -> 'Iterator[dict[str, Any]]':
        """
        Return latest request details for a DID

        Parameters
        ----------
        name:
            DID
        rse:
            Destination RSE name
        scope:
            rucio scope, defaults to None

        Raises
        -------
        exc_cls: from BaseClient._get_exception

        Returns
        -------
        request information
        """

        if scope is not None:
            path = '/'.join([self.REQUEST_BASEURL, 'history', quote_plus(scope), quote_plus(name), rse])
        url = build_url(choice(self.list_hosts), path=path)
        r = self._send_request(url, method=HTTPMethod.GET)

        if r.status_code == codes.ok:
            return next(self._load_json_data(r))
        else:
            exc_cls, exc_msg = self._get_exception(headers=r.headers, status_code=r.status_code, data=r.content)
            raise exc_cls(exc_msg)

    def list_transfer_limits(
            self
    ) -> 'Iterator[dict[str, Any]]':
        """Returns all the transfer limits

        :returns: transfer limits
        """
        path = '/'.join([self.REQUEST_BASEURL, 'transfer_limits'])
        url = build_url(choice(self.list_hosts), path=path)
        r = self._send_request(url, method=HTTPMethod.GET)

        if r.status_code == codes.ok:
            return self._load_json_data(r)
        else:
            exc_cls, exc_msg = self._get_exception(headers=r.headers, status_code=r.status_code, data=r.content)
            raise exc_cls(exc_msg)

    def set_transfer_limit(
            self,
            rse_expression: str,
            activity: Optional[str] = None,
            direction: TransferLimitDirection = TransferLimitDirection.DESTINATION,
            max_transfers: Optional[int] = None,
            volume: Optional[int] = None,
            deadline: Optional[int] = None,
            strategy: Optional[str] = None,
            transfers: Optional[int] = None,
            waitings: Optional[int] = None,
    ) -> Literal[True]:
        """Set the transfer limit for a given RSE

        :param rse_expression: RSE expression string.
        :param activity: The activity.
        :param direction: The direction in which this limit applies (source/destination)
        :param max_transfers: Maximum transfers.
        :param volume: Maximum transfer volume in bytes.
        :param deadline: Maximum waiting time in hours until a datasets gets released.
        :param strategy: defines how to handle datasets: `fifo` (each file released separately) or `grouped_fifo` (wait for the entire dataset to fit)
        :param transfers: Current number of active transfers
        :param waitings: Current number of waiting transfers

        :returns: True if the transfer limit was deleted
        """
        path = '/'.join([self.REQUEST_BASEURL, 'transfer_limits'])
        url = build_url(choice(self.list_hosts), path=path)
        data = dumps({'rse_expression': rse_expression, 'activity': activity,
                      'direction': direction.value, 'max_transfers': max_transfers,
                      'volume': volume, 'deadline': deadline, 'strategy': strategy,
                      'transfers': transfers, 'waitings': waitings})
        r = self._send_request(url, method=HTTPMethod.PUT, data=data)

        if r.status_code == codes.created:
            return True
        exc_cls, exc_msg = self._get_exception(headers=r.headers, status_code=r.status_code, data=r.content)
        raise exc_cls(exc_msg)

    def delete_transfer_limit(
            self,
            rse_expression: str,
            activity: Optional[str] = None,
            direction: TransferLimitDirection = TransferLimitDirection.DESTINATION
    ) -> Literal[True]:
        """Delete the transfer limit for a given RSE

        :param rse_expression: RSE expression string.
        :param activity: The activity.
        :param direction: The direction in which this limit applies (source/destination)

        :returns: True if the transfer limit was deleted
        """
        path = '/'.join([self.REQUEST_BASEURL, 'transfer_limits'])
        url = build_url(choice(self.list_hosts), path=path)
        data = dumps({'rse_expression': rse_expression, 'activity': activity, 'direction': direction.value})
        r = self._send_request(url, method=HTTPMethod.DELETE, data=data)

        if r.status_code == codes.ok:
            return True
        exc_cls, exc_msg = self._get_exception(headers=r.headers, status_code=r.status_code, data=r.content)
        raise exc_cls(exc_msg)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import TYPE_CHECKING, Union, cast

import flask
from flask import Flask, Response

from rucio.common.constants import HTTPMethod, TransferLimitDirection
from rucio.common.exception import AccessDenied, RequestNotFound
from rucio.common.utils import APIEncoder, render_json
from rucio.core.rse import get_rses_with_attribute_value
from rucio.db.sqla.constants import RequestState
from rucio.gateway import request
from rucio.web.rest.flaskapi.authenticated_bp import AuthenticatedBlueprint
from rucio.web.rest.flaskapi.v1.common import ErrorHandlingMethodView, check_accept_header_wrapper_flask, generate_http_error_flask, json_parameters, param_get, parse_scope_name, response_headers, try_stream

if TYPE_CHECKING:
    from collections.abc import Iterator


class RequestGet(ErrorHandlingMethodView):
    """ REST API to get requests. """

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self, scope_name, rse):
        """
        ---
        summary: Get Request
        description: "Get a request for a given DID to a destination RSE."
        tags:
          - Requests
        parameters:
        - name: scope_name
          in: path
          description: "Data identifier (scope)/(name)."
          schema:
            type: string
          style: simple
        - name: rse
          in: path
          description: "Destination rse."
          schema:
            type: string
          style: simple
        responses:
          200:
            description: "OK"
            content:
              application/json:
                schema:
                  description: "The request associated with the DID and destination RSE."
                  type: object
                  properties:
                    id:
                      description: "The id of the request."
                      type: string
                    request_type:
                      description: "The request type."
                      type: string
                      enum: ["T", "U", "D", "I", "O"]
                    scope:
                      description: "The scope of the transfer."
                      type: string
                    name:
                      description: "The name of the transfer."
                      type: string
                    did_type:
                      description: "The DID type."
                      type: string
                    dest_rse_id:
                      description: "The destination RSE id."
                      type: string
                    source_rse_id:
                      description: "The source RSE id."
                      type: string
                    attributes:
                      description: "All attributes associated with the request."
                      type: string
                    state:
                      description: "The state of the request."
                      type: string
                      enum: ["Q", "G", "S", "F", "D", "L", "N", "O", "A", "M", "U", "W", "P"]
                    external_id:
                      description: "External id of the request."
                      type: string
                    external_host:
                      description: "External host of the request."
                      type: string
                    retry_count:
                      description: "The numbers of attempted retries."
                      type: integer
                    err_msg:
                      description: "An error message if one occurred."
                      type: string
                    previous_attempt_id:
                      description: "The id of the previous attempt."
                      type: string
                    rule_id:
                      description: "The id of the associated replication rule."
                      type: string
                    activity:
                      description: "The activity of the request."
                      type: string
                    bytes:
                      description: "The size of the DID in bytes."
                      type: integer
                    md5:
                      description: "The md5 checksum of the DID to transfer."
                      type: string
                    adler32:
                      description: "The adler32 checksum of the DID to transfer."
                      type: string
                    dest_url:
                      description: "The destination url."
                      type: string
                    submitted_at:
                      description: "The time the request got submitted."
                      type: string
                    started_at:
                      description: "The time the request got started."
                      type: string
                    transferred_at:
                      description: "The time the request got transferred."
                      type: string
                    estimated_at:
                      description: "The time the request got estimated."
                      type: string
                    submitter_id:
                      description: "The id of the submitter."
                      type: string
                    estimated_stated_at:
                      description: "The estimation of the started at value."
                      type: string
                    estimated_transferred_at:
                      description: "The estimation of the transferred at value."
                      type: string
                    staging_started_at:
                      description: "The time the staging got started."
                      type: string
                    staging_finished_at:
                      description: "The time the staging got finished."
                      type: string
                    account:
                      description: "The account which issued the request."
                      type: string
                    requested_at:
                      description: "The time the request got requested."
                      type: string
                    priority:
                      description: "The priority of the request."
                      type: integer
                    transfertool:
                      description: "The transfertool used."
                      type: string
                    source_rse:
                      description: "The name of the source RSE."
                      type: string
                    dest_rse:
                      description: "The name of the destination RSE."
                      type: string
          401:
            description: "Invalid Auth Token"
          404:
            description: "Not found"
          406:
            description: "Not acceptable"
        """
        try:
            scope, name = parse_scope_name(scope_name, flask.request.environ['vo'])
        except ValueError as error:
            return generate_http_error_flask(400, error)

        try:
            request_data = request.get_request_by_did(
                scope=scope,
                name=name,
                rse=rse,
                issuer=flask.request.environ['issuer'],
                vo=flask.request.environ['vo'],
            )
            return Response(json.dumps(request_data, cls=APIEncoder), content_type='application/json')
        except RequestNotFound as error:
            return generate_http_error_flask(404, error.__class__.__name__, f'No request found for DID {scope}:{name} at RSE {rse}')


class RequestHistoryGet(ErrorHandlingMethodView):
    """ REST API to get historical requests. """

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self, scope_name, rse):
        """
        ---
        summary: Get Historical Request
        description: "List a historical request for a given DID to a destination RSE."
        tags:
          - Requests
        parameters:
        - name: scope_name
          in: path
          description: "Data identifier (scope)/(name)."
          schema:
            type: string
          style: simple
        - name: rse
          in: path
          description: "Destination rse."
          schema:
            type: string
          style: simple
        responses:
          200:
            description: "OK"
            content:
              application/json:
                schema:
                  description: "The request associated with the DID and destination RSE."
                  type: object
                  properties:
                    id:
                      description: "The id of the request."
                      type: string
                    request_type:
                      description: "The request type."
                      type: string
                      enum: ["T", "U", "D", "I", "O"]
                    scope:
                      description: "The scope of the transfer."
                      type: string
                    name:
                      description: "The name of the transfer."
                      type: string
                    did_type:
                      description: "The DID type."
                      type: string
                    dest_rse_id:
                      description: "The destination RSE id."
                      type: string
                    source_rse_id:
                      description: "The source RSE id."
                      type: string
                    attributes:
                      description: "All attributes associated with the request."
                      type: string
                    state:
                      description: "The state of the request."
                      type: string
                      enum: ["Q", "G", "S", "F", "D", "L", "N", "O", "A", "M", "U", "W", "P"]
                    external_id:
                      description: "External id of the request."
                      type: string
                    external_host:
                      description: "External host of the request."
                      type: string
                    retry_count:
                      description: "The numbers of attempted retries."
                      type: integer
                    err_msg:
                      description: "An error message if one occurred."
                      type: string
                    previous_attempt_id:
                      description: "The id of the previous attempt."
                      type: string
                    rule_id:
                      description: "The id of the associated replication rule."
                      type: string
                    activity:
                      description: "The activity of the request."
                      type: string
                    bytes:
                      description: "The size of the DID in bytes."
                      type: integer
                    md5:
                      description: "The md5 checksum of the DID to transfer."
                      type: string
                    adler32:
                      description: "The adler32 checksum of the DID to transfer."
                      type: string
                    dest_url:
                      description: "The destination url."
                      type: string
                    submitted_at:
                      description: "The time the request got submitted."
                      type: string
                    started_at:
                      description: "The time the request got started."
                      type: string
                    transferred_at:
                      description: "The time the request got transferred."
                      type: string
                    estimated_at:
                      description: "The time the request got estimated."
                      type: string
                    submitter_id:
                      description: "The id of the submitter."
                      type: string
                    estimated_stated_at:
                      description: "The estimation of the started at value."
                      type: string
                    estimated_transferred_at:
                      description: "The estimation of the transferred at value."
                      type: string
                    staging_started_at:
                      description: "The time the staging got started."
                      type: string
                    staging_finished_at:
                      description: "The time the staging got finished."
                      type: string
                    account:
                      description: "The account which issued the request."
                      type: string
                    requested_at:
                      description: "The time the request got requested."
                      type: string
                    priority:
                      description: "The priority of the request."
                      type: integer
                    transfertool:
                      description: "The transfertool used."
                      type: string
                    source_rse:
                      description: "The name of the source RSE."
                      type: string
                    dest_rse:
                      description: "The name of the destination RSE."
                      type: string
          401:
            description: "Invalid Auth Token"
          404:
            description: "Not found"
          406:
            description: "Not acceptable"
        """
        try:
            scope, name = parse_scope_name(scope_name, flask.request.environ['vo'])
        except ValueError as error:
            return generate_http_error_flask(400, error)

        try:
            request_data = request.get_request_history_by_did(
                scope=scope,
                name=name,
                rse=rse,
                issuer=flask.request.environ['issuer'],
                vo=flask.request.environ['vo'],
            )
            return Response(json.dumps(request_data, cls=APIEncoder), content_type='application/json')
        except RequestNotFound as error:
            return generate_http_error_flask(404, error.__class__.__name__, f'No request found for DID {scope}:{name} at RSE {rse}')


class RequestList(ErrorHandlingMethodView):
    """ REST API to get requests. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self):
        """
        ---
        summary: List Requests
        description: "List requests for a given source and destination RSE or site."
        tags:
          - Requests
        parameters:
        - name: src_rse
          in: query
          description: "The source rse."
          schema:
            type: array
            items:
              type: object
              required:
                - rse_id
              properties:
                rse_id:
                  description: "The id of the rse."
                  type: string
        - name: dest_rse
          in: query
          description: "The destination rse."
          schema:
            type: array
            items:
              type: object
              required:
                - rse_id
              properties:
                rse_id:
                  description: "The id of the rse."
                  type: string
        - name: src_site
          in: query
          description: "The source site."
          schema:
            type: string
        - name: dest_site
          in: query
          description: "The destination site."
          schema:
            type: string
        - name: request_states
          in: query
          description: "The accepted request states. Delimited by comma."
          schema:
            type: string
        responses:
          200:
            description: "OK"
            content:
              application/x-json-stream:
                schema:
                  description: "All requests matching the arguments. Separated by the new line character."
                  type: array
                  items:
                    description: "A request."
                    type: object
                    properties:
                      id:
                        description: "The id of the request."
                        type: string
                      request_type:
                        description: "The request type."
                        type: string
                        enum: ["T", "U", "D", "I", "O"]
                      scope:
                        description: "The scope of the transfer."
                        type: string
                      name:
                        description: "The name of the transfer."
                        type: string
                      did_type:
                        description: "The DID type."
                        type: string
                      dest_rse_id:
                        description: "The destination RSE id."
                        type: string
                      source_rse_id:
                        description: "The source RSE id."
                        type: string
                      attributes:
                        description: "All attributes associated with the request."
                        type: string
                      state:
                        description: "The state of the request."
                        type: string
                        enum: ["Q", "G", "S", "F", "D", "L", "N", "O", "A", "M", "U", "W", "P"]
                      external_id:
                        description: "External id of the request."
                        type: string
                      external_host:
                        description: "External host of the request."
                        type: string
                      retry_count:
                        description: "The numbers of attempted retries."
                        type: integer
                      err_msg:
                        description: "An error message if one occurred."
                        type: string
                      previous_attempt_id:
                        description: "The id of the previous attempt."
                        type: string
                      rule_id:
                        description: "The id of the associated replication rule."
                        type: string
                      activity:
                        description: "The activity of the request."
                        type: string
                      bytes:
                        description: "The size of the DID in bytes."
                        type: integer
                      md5:
                        description: "The md5 checksum of the DID to transfer."
                        type: string
                      adler32:
                        description: "The adler32 checksum of the DID to transfer."
                        type: string
                      dest_url:
                        description: "The destination url."
                        type: string
                      submitted_at:
                        description: "The time the request got submitted."
                        type: string
                      started_at:
                        description: "The time the request got started."
                        type: string
                      transferred_at:
                        description: "The time the request got transferred."
                        type: string
                      estimated_at:
                        description: "The time the request got estimated."
                        type: string
                      submitter_id:
                        description: "The id of the submitter."
                        type: string
                      estimated_stated_at:
                        description: "The estimation of the started at value."
                        type: string
                      estimated_transferred_at:
                        description: "The estimation of the transferred at value."
                        type: string
                      staging_started_at:
                        description: "The time the staging got started."
                        type: string
                      staging_finished_at:
                        description: "The time the staging got finished."
                        type: string
                      account:
                        description: "The account which issued the request."
                        type: string
                      requested_at:
                        description: "The time the request got requested."
                        type: string
                      priority:
                        description: "The priority of the request."
                        type: integer
                      transfertool:
                        description: "The transfertool used."
                        type: string
                      source_rse:
                        description: "The name of the source RSE."
                        type: string
                      dest_rse:
                        description: "The name of the destination RSE."
                        type: string
          401:
            description: "Invalid Auth Token"
          404:
            description: "Not found"
          406:
            description: "Not acceptable"
        """
        src_rse = flask.request.args.get('src_rse', default=None)
        dst_rse = flask.request.args.get('dst_rse', default=None)
        src_site = flask.request.args.get('src_site', default=None)
        dst_site = flask.request.args.get('dst_site', default=None)
        request_states = flask.request.args.get('request_states', default=None)

        if not request_states:
            return generate_http_error_flask(400, 'MissingParameter', 'Request state is missing')
        if src_rse and not dst_rse:
            return generate_http_error_flask(400, 'MissingParameter', 'Destination RSE is missing')
        elif dst_rse and not src_rse:
            return generate_http_error_flask(400, 'MissingParameter', 'Source RSE is missing')
        elif src_site and not dst_site:
            return generate_http_error_flask(400, 'MissingParameter', 'Destination site is missing')
        elif dst_site and not src_site:
            return generate_http_error_flask(400, 'MissingParameter', 'Source site is missing')

        try:
            states = [RequestState(state) for state in request_states.split(',')]
        except ValueError:
            return generate_http_error_flask(400, 'Invalid', 'Request state value is invalid')

        src_rses = []
        dst_rses = []
        if src_site:
            src_rses = get_rses_with_attribute_value(key='site', value=src_site, vo=flask.request.environ['vo'])
            if not src_rses:
                return generate_http_error_flask(404, 'NotFound', f'Could not resolve site name {src_site} to RSE')
            src_rses = [rse['rse_name'] for rse in src_rses]
            dst_rses = get_rses_with_attribute_value(key='site', value=dst_site, vo=flask.request.environ['vo'])
            if not dst_rses:
                return generate_http_error_flask(404, 'NotFound', f'Could not resolve site name {dst_site} to RSE')
            dst_rses = [rse['rse_name'] for rse in dst_rses]
        else:
            dst_rses = [dst_rse]
            src_rses = [src_rse]

        # Manual cast to list[str] as static code analysis erroneously sees these as list[Optional[str]]
        src_rses = cast("list[str]", src_rses)
        dst_rses = cast("list[str]", dst_rses)

        def generate(issuer, vo):
            for result in request.list_requests(src_rses, dst_rses, states, issuer=issuer, vo=vo):
                yield render_json(**result) + '\n'

        return try_stream(generate(issuer=flask.request.environ['issuer'], vo=flask.request.environ['vo']))


class RequestHistoryList(ErrorHandlingMethodView):
    """ REST API to get requests. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self):
        """
        ---
        summary: List Historic Requests
        description: "List historical requests for a given source and destination RSE or site."
        tags:
          - Requests
        parameters:
        - name: src_rse
          in: query
          description: "The source rse."
          schema:
            type: array
            items:
              type: object
              required:
                - rse_id
              properties:
                rse_id:
                  description: "The id of the rse."
                  type: string
        - name: dest_rse
          in: query
          description: "The destination rse."
          schema:
            type: array
            items:
              type: object
              required:
                - rse_id
              properties:
                rse_id:
                  description: "The id of the rse."
                  type: string
        - name: src_site
          in: query
          description: "The source site."
          schema:
            type: string
        - name: dest_site
          in: query
          description: "The destination site."
          schema:
            type: string
        - name: request_states
          in: query
          description: "The accepted request states. Delimited by comma."
          schema:
            type: string
        - name: offset
          in: query
          description: "The offset of the list."
          schema:
            type: integer
            default: 0
        - name: limit
          in: query
          description: "The maximum number of items to return."
          schema:
            type: integer
            default: 100
        responses:
          200:
            description: "OK"
            content:
              application/x-json-stream:
                schema:
                  description: "All requests matching the arguments. Separated by a new line character."
                  type: array
                  items:
                    description: "A request."
                    type: object
                    properties:
                      id:
                        description: "The id of the request."
                        type: string
                      request_type:
                        description: "The request type."
                        type: string
                        enum: ["T", "U", "D", "I", "O"]
                      scope:
                        description: "The scope of the transfer."
                        type: string
                      name:
                        description: "The name of the transfer."
                        type: string
                      did_type:
                        description: "The DID type."
                        type: string
                      dest_rse_id:
                        description: "The destination RSE id."
                        type: string
                      source_rse_id:
                        description: "The source RSE id."
                        type: string
                      attributes:
                        description: "All attributes associated with the request."
                        type: string
                      state:
                        description: "The state of the request."
                        type: string
                        enum: ["Q", "G", "S", "F", "D", "L", "N", "O", "A", "M", "U", "W", "P"]
                      external_id:
                        description: "External id of the request."
                        type: string
                      external_host:
                        description: "External host of the request."
                        type: string
                      retry_count:
                        description: "The numbers of attempted retries."
                        type: integer
                      err_msg:
                        description: "An error message if one occurred."
                        type: string
                      previous_attempt_id:
                        description: "The id of the previous attempt."
                        type: string
                      rule_id:
                        description: "The id of the associated replication rule."
                        type: string
                      activity:
                        description: "The activity of the request."
                        type: string
                      bytes:
                        description: "The size of the DID in bytes."
                        type: integer
                      md5:
                        description: "The md5 checksum of the DID to transfer."
                        type: string
                      adler32:
                        description: "The adler32 checksum of the DID to transfer."
                        type: string
                      dest_url:
                        description: "The destination url."
                        type: string
                      submitted_at:
                        description: "The time the request got submitted."
                        type: string
                      started_at:
                        description: "The time the request got started."
                        type: string
                      transferred_at:
                        description: "The time the request got transferred."
                        type: string
                      estimated_at:
                        description: "The time the request got estimated."
                        type: string
                      submitter_id:
                        description: "The id of the submitter."
                        type: string
                      estimated_stated_at:
                        description: "The estimation of the started at value."
                        type: string
                      estimated_transferred_at:
                        description: "The estimation of the transferred at value."
                        type: string
                      staging_started_at:
                        description: "The time the staging got started."
                        type: string
                      staging_finished_at:
                        description: "The time the staging got finished."
                        type: string
                      account:
                        description: "The account which issued the request."
                        type: string
                      requested_at:
                        description: "The time the request got requested."
                        type: string
                      priority:
                        description: "The priority of the request."
                        type: integer
                      transfertool:
                        description: "The transfertool used."
                        type: string
                      source_rse:
                        description: "The name of the source RSE."
                        type: string
                      dest_rse:
                        description: "The name of the destination RSE."
                        type: string
          401:
            description: "Invalid Auth Token"
          404:
            description: "Not found"
          406:
            description: "Not acceptable"
        """
        src_rse = flask.request.args.get('src_rse', default=None)
        dst_rse = flask.request.args.get('dst_rse', default=None)
        src_site = flask.request.args.get('src_site', default=None)
        dst_site = flask.request.args.get('dst_site', default=None)
        request_states = flask.request.args.get('request_states', default=None)
        offset = flask.request.args.get('offset', type=int, default=0)
        limit = flask.request.args.get('limit', type=int, default=100)

        if not request_states:
            return generate_http_error_flask(400, 'MissingParameter', 'Request state is missing')
        if src_rse and not dst_rse:
            return generate_http_error_flask(400, 'MissingParameter', 'Destination RSE is missing')
        elif dst_rse and not src_rse:
            return generate_http_error_flask(400, 'MissingParameter', 'Source RSE is missing')
        elif src_site and not dst_site:
            return generate_http_error_flask(400, 'MissingParameter', 'Destination site is missing')
        elif dst_site and not src_site:
            return generate_http_error_flask(400, 'MissingParameter', 'Source site is missing')

        try:
            states = [RequestState(state) for state in request_states.split(',')]
        except ValueError:
            return generate_http_error_flask(400, 'Invalid', 'Request state value is invalid')

        src_rses = []
        dst_rses = []
        if src_site:
            src_rses = get_rses_with_attribute_value(key='site', value=src_site, vo=flask.request.environ['vo'])
            if not src_rses:
                return generate_http_error_flask(404, 'NotFound', f'Could not resolve site name {src_site} to RSE')
            src_rses = [rse['rse_name'] for rse in src_rses]
            dst_rses = get_rses_with_attribute_value(key='site', value=dst_site, vo=flask.request.environ['vo'])
            if not dst_rses:
                return generate_http_error_flask(404, 'NotFound', f'Could not resolve site name {dst_site} to RSE')
            dst_rses = [rse['rse_name'] for rse in dst_rses]
        else:
            dst_rses = [dst_rse]
            src_rses = [src_rse]

        # Manual cast to list[str] as static code analysis erroneously sees these as list[Optional[str]]
        src_rses = cast("list[str]", src_rses)
        dst_rses = cast("list[str]", dst_rses)

        def generate(issuer, vo):
            for result in request.list_requests_history(src_rses, dst_rses, states, issuer=issuer, vo=vo, offset=offset, limit=limit):
                yield render_json(**result) + '\n'

        return try_stream(generate(issuer=flask.request.environ['issuer'], vo=flask.request.environ['vo']))


class RequestMetricsGet(ErrorHandlingMethodView):
    """ REST API to get request stats. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self):
        """
        ---
        summary: Get Request Statistics
        description: "Get statistics of requests grouped by source, destination, and activity."
        tags:
          - Requests
        parameters:
        - name: dest_rse
          in: query
          description: "The destination RSE name"
          schema:
            type: string
        - name: source_rse
          in: query
          description: "The source RSE name"
          schema:
            type: string
        - name: activity
          in: query
          description: "The activity"
          schema:
            type: string
        - name: group_by_rse_attribute
          in: query
          description: "The parameter to group the RSEs by."
          schema:
            type: string
        responses:
          200:
            description: "OK"
            content:
              application/x-json-stream:
                schema:
                  description: "Statistics of requests by source, destination, and activity."
                  type: array
                  items:
                    description: "Statistics of the request group for a given (source, destination, activity) tuple."
                    type: object
                    properties:
                      src_rse:
                        type: string
                        description: "The name of this links source RSE"
                      dst_rse:
                        type: string
                        description: "The name of this links destination RSE"
                      distance:
                        type: integer
                        description: "The distance between the source and destination RSE"
                      files:
                        type: object
                        properties:
                          done-total-1h:
                            type: integer
                            description: "The total number of files successfully transferred in the last 1 hour"
                          done-total-6h:
                            type: integer
                            description: "The total number of files successfully transferred in the last 6 hours"
                          failed-total-1h:
                            type: integer
                            description: "The total number of transfer failures in the last 1 hour"
                          failed-total-6h:
                            type: integer
                            description: "The total number of transfer failures in the last 6 hours"
                          queued-total:
                            type: integer
                            description: "The total number of files queued in rucio"
                          queued:
                            type: object
                            description: "Per-activity number of queued files"
                            additionalProperties:
                              type: integer
                          done:
                            type: object
                            additionalProperties:
                              type: object
                              properties:
                                1h:
                                  type: integer
                                6h:
                                  type: integer
                          failed:
                            type: object
                            description: "Per-activity number of transfer failures in the last 1 and 6 hours"
                            additionalProperties:
                              type: object
                              properties:
                                1h:
                                  type: integer
                                6h:
                                  type: integer
                      bytes:
                        type: object
                        properties:
                          done-total-1h:
                            type: integer
                            description: "The total number of bytes successfully transferred in the last 1 hour"
                          done-total-6h:
                            type: integer
                            description: "The total number of bytes successfully transferred in the last 6 hours"
                          queued-total:
                            type: integer
                            description: "The total number of bytes queued to be transferred by rucio"
                          queued:
                            type: object
                            description: "Per-activity amount of queued bytes"
                            additionalProperties:
                              type: integer
                          done:
                            type: object
                            description: "Per-activity number of transferred bytes in the last 1 and 6 hours"
                            additionalProperties:
                              type: object
                              properties:
                                1h:
                                  type: integer
                                6h:
                                  type: integer
                    required:
                      - distance
                      - src_rse
                      - dst_rse
          401:
            description: "Invalid Auth Token"
        """
        dst_rse = flask.request.args.get('dst_rse', default=None)
        src_rse = flask.request.args.get('src_rse', default=None)
        activity = flask.request.args.get('activity', default=None)
        group_by_rse_attribute = flask.request.args.get('group_by_rse_attribute', default=None)
        format = flask.request.args.get('format', default=None)

        metrics = request.get_request_metrics(
            dst_rse=dst_rse,
            src_rse=src_rse,
            activity=activity,
            group_by_rse_attribute=group_by_rse_attribute,
            issuer=flask.request.environ['issuer'],
            vo=flask.request.environ['vo']
        )

        if format == 'panda':
            return Response(json.dumps(metrics, cls=APIEncoder), content_type='application/json')

        def generate() -> "Iterator[str]":
            for result in metrics.values():
                yield render_json(**result) + '\n'
        return try_stream(generate())


class TransferLimits(ErrorHandlingMethodView):
    """ REST API to get, set or delete transfer limits. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self) -> flask.Response:
        """
        ---
        summary: Get Transfer Limits
        description: "Get all the transfer limits."
        tags:
          - Requests
        responses:
          200:
            description: "OK"
            content:
              application/x-json-stream:
                schema:
                  description: "All the transfer limits"
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        description: "The transfer limit id."
                        type: string
                      rse_expression:
                        description: "The RSE expression for which the limit applies."
                        type: string
                      direction:
                        description: "The direction in which this limit applies (source/destination)"
                        type: string
                      max_transfers:
                        description: "Maximum number of transfers allowed."
                        type: integer
                      volume:
                        description: "Maximum transfer volume in bytes."
                        type: integer
                      deadline:
                        description: "Maximum waiting time in hours until a dataset gets released."
                        type: integer
                      strategy:
                        description: "Defines how to handle datasets: `fifo` (each file released separately) or `grouped_fifo` (wait for the entire dataset to fit)"
                        type: string
                      transfers:
                        description: "Current number of active transfers"
                        type: integer
                      waitings:
                        description: "Current number of waiting transfers"
                        type: integer
                      updated_at:
                        description: "Datetime of the last update."
                        type: string
                      created_at:
                        description: "Datetime of the creation of the transfer limit."
                        type: string
          401:
            description: "Invalid Auth Token"
        """
        transfer_limits = request.list_transfer_limits(issuer=flask.request.environ['issuer'], vo=flask.request.environ['vo'])

        def generate() -> "Iterator[str]":
            for limit in transfer_limits:
                yield json.dumps(limit, cls=APIEncoder) + '\n'
        return try_stream(generate())

    def put(self) -> Union[flask.Response, tuple[str, int]]:
        """
        ---
        summary: Set Transfer Limit
        description: "Create or update a transfer limit for a specific RSE expression and activity."
        tags:
          - Requests
        requestBody:
          content:
            application/json:
              schema:
                type: object
                required:
                  - rse_expression
                  - max_transfers
                properties:
                  rse_expression:
                    type: string
                    description: "The RSE expression for which the transfer limit is being set."
                  activity:
                    type: string
                    description: "The activity to which the transfer limit applies."
                  max_transfers:
                    type: integer
                    description: "The maximum number of transfers allowed."
                  direction:
                    type: string
                    description: "The direction of the transfer limit (source or destination)."
                    enum: ["SOURCE", "DESTINATION"]
                    default: "DESTINATION"
                  volume:
                    type: integer
                    description: "The maximum transfer volume in bytes."
                  deadline:
                    type: integer
                    description: "The maximum waiting time in hours until a dataset is released."
                  strategy:
                    type: string
                    description: "The strategy for handling datasets (e.g., `fifo` or `grouped_fifo`)."
                  transfers:
                    type: integer
                    description: "The current number of active transfers."
                  waitings:
                    type: integer
                    description: "The current number of waiting transfers."
        responses:
          201:
            description: "Transfer limit set successfully."
          400:
            description: "Invalid input data."
          401:
            description: "Invalid Auth Token."
          500:
            description: "Internal server error."
        """
        parameters = json_parameters()
        rse_expression = param_get(parameters, 'rse_expression')
        max_transfers = param_get(parameters, 'max_transfers')

        try:
            request.set_transfer_limit(
                rse_expression=rse_expression,
                max_transfers=max_transfers,
                activity=param_get(parameters, 'activity', default=None),
                direction=param_get(parameters, 'direction', default=TransferLimitDirection.DESTINATION),
                volume=param_get(parameters, 'volume', default=None),
                deadline=param_get(parameters, 'deadline', default=None),
                strategy=param_get(parameters, 'strategy', default=None),
                transfers=param_get(parameters, 'transfers', default=None),
                waitings=param_get(parameters, 'waitings', default=None),
                issuer=flask.request.environ['issuer'],
                vo=flask.request.environ['vo']
            )
        except AccessDenied as error:
            return generate_http_error_flask(401, error)

        return '', 201

    def delete(self) -> Union[flask.Response, tuple[str, int]]:
        """
        ---
        summary: Delete Transfer Limit
        description: "Delete a transfer limit for an RSE expression."
        tags:
          - Requests
        parameters:
          - name: rse_expression
            in: query
            description: "The RSE expression to delete the limit for."
            required: true
            schema:
              type: string
        responses:
          200:
            description: "Transfer limit deleted successfully."
          400:
            description: "Invalid input data."
          401:
            description: "Invalid Auth Token."
          500:
            description: "Internal server error."
        """
        parameters = json_parameters()
        rse_expression = param_get(parameters, 'rse_expression')

        try:
            request.delete_transfer_limit(
                rse_expression=rse_expression,
                activity=param_get(parameters, 'activity', default=None),
                direction=param_get(parameters, 'direction', default=TransferLimitDirection.DESTINATION),
                issuer=flask.request.environ['issuer'],
                vo=flask.request.environ['vo']
            )
        except AccessDenied as error:
            return generate_http_error_flask(401, error)

        return '', 200


def blueprint() -> AuthenticatedBlueprint:
    bp = AuthenticatedBlueprint('requests', __name__, url_prefix='/requests')

    request_get_view = RequestGet.as_view('request_get')
    bp.add_url_rule('/<path:scope_name>/<rse>', view_func=request_get_view, methods=[HTTPMethod.GET.value])
    request_history_get_view = RequestHistoryGet.as_view('request_history_get')
    bp.add_url_rule('/history/<path:scope_name>/<rse>', view_func=request_history_get_view, methods=[HTTPMethod.GET.value])
    request_list_view = RequestList.as_view('request_list')
    bp.add_url_rule('/list', view_func=request_list_view, methods=[HTTPMethod.GET.value])
    request_history_list_view = RequestHistoryList.as_view('request_history_list')
    bp.add_url_rule('/history/list', view_func=request_history_list_view, methods=[HTTPMethod.GET.value])
    request_metrics_view = RequestMetricsGet.as_view('request_metrics_get')
    bp.add_url_rule('/metrics', view_func=request_metrics_view, methods=[HTTPMethod.GET.value])
    transfer_limits_view = TransferLimits.as_view('transfer_limits_get')
    bp.add_url_rule('/transfer_limits', view_func=transfer_limits_view, methods=[HTTPMethod.GET.value, HTTPMethod.PUT.value, HTTPMethod.DELETE.value])

    bp.after_request(response_headers)
    return bp


def make_doc():
    """ Only used for sphinx documentation """
    doc_app = Flask(__name__)
    doc_app.register_blueprint(blueprint())
    return doc_app



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from sqlalchemy import and_, delete, select
from sqlalchemy.exc import NoResultFound

from rucio.client.didclient import DIDClient
from rucio.client.replicaclient import ReplicaClient
from rucio.client.ruleclient import RuleClient
from rucio.common.exception import InvalidObject
from rucio.core.did import add_dids, attach_dids
from rucio.core.replica import add_replicas, delete_replicas, get_cleaned_updated_collection_replicas, list_datasets_per_rse, update_collection_replica
from rucio.core.rse import add_protocol, add_rse, del_rse, get_rse_id
from rucio.db.sqla import constants, models
from rucio.db.sqla.constants import ReplicaState
from rucio.tests.common import did_name_generator, rse_name_generator


class TestDatasetReplicaClient:

    @pytest.mark.noparallel(reason='uses pre-defined RSE')
    def test_list_dataset_replicas(self):
        """ REPLICA (CLIENT): List dataset replicas."""
        activity = "Staging"
        replica_client = ReplicaClient()
        rule_client = RuleClient()
        did_client = DIDClient()
        scope = 'mock'
        dataset = did_name_generator('dataset')

        did_client.add_dataset(scope=scope, name=dataset)
        rule_client.add_replication_rule(dids=[{'scope': scope, 'name': dataset}],
                                         account='root', copies=1, rse_expression='MOCK',
                                         grouping='DATASET', activity=activity)
        replicas = [r for r in replica_client.list_dataset_replicas(scope=scope, name=dataset)]
        assert len(replicas) == 1

    @pytest.mark.noparallel(reason='uses pre-defined RSE')
    def test_list_dataset_replicas_bulk(self):
        """ REPLICA (CLIENT): List dataset replicas bulk."""
        activity = "Staging"
        replica_client = ReplicaClient()
        rule_client = RuleClient()
        did_client = DIDClient()
        scope = 'mock'
        did1 = {'scope': scope, 'name': did_name_generator('dataset')}
        did_client.add_dataset(**did1)
        did2 = {'scope': scope, 'name': did_name_generator('dataset')}
        did_client.add_dataset(**did2)
        dids = [did1, did2]
        rule_client.add_replication_rule(dids=dids,
                                         account='root', copies=1, rse_expression='MOCK',
                                         grouping='DATASET', activity=activity)

        with pytest.raises(InvalidObject):
            replica_client.list_dataset_replicas_bulk(dids=[{'type': "I'm Different"}])

        replicas = list(replica_client.list_dataset_replicas_bulk(dids=dids))

        assert len(replicas) == 2
        for did in dids:
            def replica_contains_did(rep):
                return all(map(lambda k: k in rep and did[k] == rep[k], did))
            assert any(map(replica_contains_did, replicas)), "%s must be in returned replicas" % (did, )

    @pytest.mark.noparallel(reason='uses pre-defined RSE')
    def test_list_datasets_per_rse(self, vo, mock_scope):
        """ REPLICA (CLIENT): List datasets in RSE."""
        activity = "Staging"
        rule_client = RuleClient()
        did_client = DIDClient()
        dataset = did_name_generator('dataset')
        pattern = dataset[:-5] + '*'

        did_client.add_dataset(scope=mock_scope.external, name=dataset)
        rule_client.add_replication_rule(dids=[{'scope': mock_scope.external, 'name': dataset}],
                                         account='root', copies=1, rse_expression='MOCK',
                                         grouping='DATASET', activity=activity)
        replicas = [r for r in list_datasets_per_rse(rse_id=get_rse_id(rse='MOCK', vo=vo),
                                                     filters={'scope': mock_scope, 'name': pattern})]
        assert replicas != []

    def test_list_dataset_replicas_archive(self, vo, did_client):
        """ REPLICA (CLIENT): List dataset replicas with archives. """

        activity = "Staging"
        replica_client = ReplicaClient()
        did_client = DIDClient()
        rule_client = RuleClient()

        scope = 'mock'

        rse = 'APERTURE_%s' % rse_name_generator()
        rse_id = add_rse(rse, vo=vo)
        add_protocol(rse_id=rse_id, parameter={'scheme': 'root',
                                               'hostname': 'root.aperture.com',
                                               'port': 1409,
                                               'prefix': '//test/chamber/',
                                               'impl': 'rucio.rse.protocols.xrootd.Default',
                                               'domains': {
                                                   'lan': {'read': 1, 'write': 1, 'delete': 1},
                                                   'wan': {'read': 1, 'write': 1, 'delete': 1}}})

        rse2 = 'BLACKMESA_%s' % rse_name_generator()
        rse2_id = add_rse(rse2, vo=vo)
        add_protocol(rse_id=rse2_id, parameter={'scheme': 'root',
                                                'hostname': 'root.blackmesa.com',
                                                'port': 1409,
                                                'prefix': '//underground/facility',
                                                'impl': 'rucio.rse.protocols.xrootd.Default',
                                                'domains': {
                                                    'lan': {'read': 1, 'write': 1, 'delete': 1},
                                                    'wan': {'read': 1, 'write': 1, 'delete': 1}}})

        # register archive
        archive = {'scope': scope, 'name': '%sanother.zip' % did_name_generator('file'),
                   'type': 'FILE', 'bytes': 2596, 'adler32': 'deedbeaf'}
        replica_client.add_replicas(rse=rse, files=[archive])
        replica_client.add_replicas(rse=rse2, files=[archive])

        archived_files = [{'scope': scope, 'name': '%szippedfile-%i' % (did_name_generator('file'), i), 'type': 'FILE',
                           'bytes': 4322, 'adler32': 'deaddead'} for i in range(2)]
        replica_client.add_replicas(rse=rse2, files=archived_files)
        did_client.add_files_to_archive(scope=scope, name=archive['name'], files=archived_files)

        dataset_name = did_name_generator('dataset')
        did_client.add_dataset(scope=scope, name=dataset_name)
        did_client.attach_dids(scope=scope, name=dataset_name, dids=archived_files)
        rule_client.add_replication_rule(dids=[{'scope': scope, 'name': dataset_name}],
                                         account='root', copies=1, rse_expression=rse,
                                         grouping='DATASET', activity=activity)

        res = [r for r in replica_client.list_dataset_replicas(scope=scope,
                                                               name=dataset_name)]
        assert len(res) == 1
        assert res[0]['state'] == 'UNAVAILABLE'

        res = [r for r in replica_client.list_dataset_replicas(scope=scope,
                                                               name=dataset_name,
                                                               deep=True)]

        assert len(res) == 3
        assert res[0]['state'] == 'AVAILABLE'
        assert res[1]['state'] == 'AVAILABLE'
        assert res[2]['state'] == 'AVAILABLE'

        # Update replica state of the archive to TEMPORARY_UNAVAILABLE
        replica_client.update_replicas_states(
            rse=rse,
            files=[{'scope': archive['scope'], 'name': archive['name'], 'state': 'T'}]
        )

        # Running list_dataset_replicas with deep=True should now show no replicas
        # from RSE since the archive is TEMPORARY_UNAVAILABLE
        res = [r for r in replica_client.list_dataset_replicas(scope=scope,
                                                               name=dataset_name,
                                                               deep=True)]

        # All replicas from RSE should be gone. The number of remaining replicas should be 2.
        assert len(res) == 2
        assert rse not in [r['rse'] for r in res]

        del_rse(rse_id)


@pytest.mark.noparallel(reason='uses pre-defined RSEs, truncates table(s) updated_col_rep')
class TestDatasetReplicaUpdate:

    def test_clean_and_get_collection_replica_updates(self, mock_scope, rse_factory, root_account, db_session):
        """ REPLICA (CORE): Get cleaned update requests for collection replicas. """
        rse, rse_id = rse_factory.make_mock_rse()
        dataset_name_with_collection_replica = did_name_generator('dataset')
        dataset_name_without_collection_replica = did_name_generator('dataset')
        add_dids(dids=[{'name': dataset_name_without_collection_replica, 'scope': mock_scope, 'type': constants.DIDType.DATASET},
                       {'name': dataset_name_with_collection_replica, 'scope': mock_scope, 'type': constants.DIDType.DATASET}], account=root_account, session=db_session)
        stmt = delete(models.UpdatedCollectionReplica)
        db_session.execute(stmt)
        db_session.commit()  # pylint: disable=no-member

        # setup test data - 4 without corresponding replica, 4 duplicates and 2 correct
        models.CollectionReplica(rse_id=rse_id, scope=mock_scope, bytes=10, length=0, available_replicas_cnt=0, state=constants.ReplicaState.AVAILABLE, name=dataset_name_with_collection_replica, did_type=constants.DIDType.DATASET)\
            .save(session=db_session)
        models.UpdatedCollectionReplica(rse_id=rse_id, scope=mock_scope, name=dataset_name_with_collection_replica, did_type=constants.DIDType.DATASET).save(session=db_session)
        models.UpdatedCollectionReplica(rse_id=rse_id, scope=mock_scope, name=dataset_name_with_collection_replica, did_type=constants.DIDType.DATASET).save(session=db_session)
        models.UpdatedCollectionReplica(rse_id=rse_id, scope=mock_scope, name=dataset_name_with_collection_replica, did_type=constants.DIDType.DATASET).save(session=db_session)
        models.UpdatedCollectionReplica(rse_id=rse_id, scope=mock_scope, name=dataset_name_without_collection_replica, did_type=constants.DIDType.DATASET).save(session=db_session)
        models.UpdatedCollectionReplica(rse_id=rse_id, scope=mock_scope, name=dataset_name_without_collection_replica, did_type=constants.DIDType.DATASET).save(session=db_session)
        models.UpdatedCollectionReplica(scope=mock_scope, name=dataset_name_without_collection_replica, did_type=constants.DIDType.DATASET).save(session=db_session)
        models.UpdatedCollectionReplica(scope=mock_scope, name=dataset_name_without_collection_replica, did_type=constants.DIDType.DATASET).save(session=db_session)
        models.UpdatedCollectionReplica(scope=mock_scope, name=dataset_name_with_collection_replica, did_type=constants.DIDType.DATASET).save(session=db_session)
        models.UpdatedCollectionReplica(scope=mock_scope, name=dataset_name_with_collection_replica, did_type=constants.DIDType.DATASET).save(session=db_session)
        models.UpdatedCollectionReplica(scope=mock_scope, name=dataset_name_with_collection_replica, did_type=constants.DIDType.DATASET).save(session=db_session)

        cleaned_collection_replica_updates = get_cleaned_updated_collection_replicas(total_workers=0, worker_number=0, session=db_session)
        assert len(cleaned_collection_replica_updates) == 2
        for update_request in cleaned_collection_replica_updates:
            stmt = select(
                models.UpdatedCollectionReplica
            ).where(
                models.UpdatedCollectionReplica.id == update_request['id']
            )
            update_request = db_session.execute(stmt).scalar_one()
            assert update_request.scope == mock_scope
            assert update_request.name in (dataset_name_with_collection_replica, dataset_name_without_collection_replica)

    def test_update_collection_replica(self, mock_scope, rse_factory, root_account, db_session):
        """ REPLICA (CORE): Update collection replicas from update requests. """
        rse, rse_id = rse_factory.make_mock_rse()
        rse2, rse2_id = rse_factory.make_mock_rse()
        file_size = 2
        files = [{'name': 'file_%s' % did_name_generator('file'), 'scope': mock_scope, 'bytes': file_size} for i in range(0, 2)]
        dataset_name = 'dataset_test_%s' % did_name_generator('dataset')
        add_replicas(rse_id=rse_id, files=files, account=root_account, session=db_session)
        add_dids([{'scope': mock_scope, 'name': dataset_name, 'type': constants.DIDType.DATASET}], account=root_account, session=db_session)
        attach_dids(scope=mock_scope, name=dataset_name, dids=files, account=root_account, session=db_session)
        models.CollectionReplica(rse_id=rse_id, scope=mock_scope, state=constants.ReplicaState.AVAILABLE, name=dataset_name, did_type=constants.DIDType.DATASET, bytes=len(files) * file_size, length=len(files), available_replicas_cnt=0)\
              .save(session=db_session)

        # Update request with rse id
        # First update -> dataset replica should be available
        models.UpdatedCollectionReplica(rse_id=rse_id, scope=mock_scope, name=dataset_name, did_type=constants.DIDType.DATASET).save(session=db_session)
        stmt = select(
            models.UpdatedCollectionReplica
        ).where(
            and_(models.UpdatedCollectionReplica.rse_id == rse_id,
                 models.UpdatedCollectionReplica.scope == mock_scope,
                 models.UpdatedCollectionReplica.name == dataset_name)
        )
        update_request = db_session.execute(stmt).scalar_one()
        update_collection_replica(update_request=update_request.to_dict(), session=db_session)
        stmt = select(
            models.UpdatedCollectionReplica
        ).where(
            models.UpdatedCollectionReplica.id == update_request.id
        )
        update_request = db_session.execute(stmt).scalar()
        assert update_request is None
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == len(files) * file_size
        assert dataset_replica['length'] == len(files)
        assert dataset_replica['available_bytes'] == len(files) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files)
        assert dataset_replica['state'] == ReplicaState.AVAILABLE

        # Delete one file replica -> dataset replica should be unavailable
        delete_replicas(rse_id=rse_id, files=[files[0]], session=db_session)
        db_session.commit()
        stmt = select(
            models.UpdatedCollectionReplica
        ).where(
            and_(models.UpdatedCollectionReplica.rse_id == rse_id,
                 models.UpdatedCollectionReplica.scope == mock_scope,
                 models.UpdatedCollectionReplica.name == dataset_name)
        )
        update_request = db_session.execute(stmt).scalar_one()
        update_collection_replica(update_request=update_request.to_dict(), session=db_session)
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == len(files) * file_size
        assert dataset_replica['length'] == len(files)
        assert dataset_replica['available_bytes'] == (len(files) - 1) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files) - 1
        assert dataset_replica['state'] == ReplicaState.UNAVAILABLE

        # Add one file replica -> dataset replica should be available again
        add_replicas(rse_id=rse_id, files=[files[0]], account=root_account, session=db_session)
        attach_dids(scope=mock_scope, name=dataset_name, dids=[files[0]], account=root_account, session=db_session)
        models.UpdatedCollectionReplica(rse_id=rse_id, scope=mock_scope, name=dataset_name, did_type=constants.DIDType.DATASET).save(session=db_session)
        stmt = select(
            models.UpdatedCollectionReplica
        ).where(
            and_(models.UpdatedCollectionReplica.rse_id == rse_id,
                 models.UpdatedCollectionReplica.scope == mock_scope,
                 models.UpdatedCollectionReplica.name == dataset_name)
        )
        update_request = db_session.execute(stmt).scalar_one()
        update_collection_replica(update_request=update_request.to_dict(), session=db_session)
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == len(files) * file_size
        assert dataset_replica['length'] == len(files)
        assert dataset_replica['available_bytes'] == len(files) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files)
        assert dataset_replica['state'] == ReplicaState.AVAILABLE

        # Old behaviour, open empty datasets are not deleted
        # Delete all file replicas -> dataset replica should be deleted
        delete_replicas(rse_id=rse_id, files=files, session=db_session)
        db_session.commit()
        with pytest.raises(NoResultFound):
            update_collection_replica(update_request=update_request.to_dict(), session=db_session)

        # Update request without rse_id - using two replicas per file -> total 4 replicas
        dataset_name = 'dataset_test_%s' % did_name_generator('dataset')
        add_dids([{'scope': mock_scope, 'name': dataset_name, 'type': constants.DIDType.DATASET}], account=root_account, session=db_session)
        add_replicas(rse_id=rse_id, files=files, account=root_account, session=db_session)
        add_replicas(rse_id=rse2_id, files=files, account=root_account, session=db_session)
        attach_dids(scope=mock_scope, name=dataset_name, dids=files, account=root_account, session=db_session)
        models.CollectionReplica(rse_id=rse_id, scope=mock_scope, name=dataset_name, state=constants.ReplicaState.UNAVAILABLE, did_type=constants.DIDType.DATASET, bytes=len(files) * file_size, length=len(files)).save(session=db_session)
        models.CollectionReplica(rse_id=rse2_id, scope=mock_scope, name=dataset_name, state=constants.ReplicaState.UNAVAILABLE, did_type=constants.DIDType.DATASET, bytes=len(files) * file_size, length=len(files)).save(session=db_session)

        # First update -> replicas should be available
        models.UpdatedCollectionReplica(scope=mock_scope, name=dataset_name, did_type=constants.DIDType.DATASET).save(session=db_session)
        stmt = select(
            models.UpdatedCollectionReplica
        ).where(
            and_(models.UpdatedCollectionReplica.scope == mock_scope,
                 models.UpdatedCollectionReplica.name == dataset_name)
        )
        update_request = db_session.execute(stmt).scalar_one()
        update_collection_replica(update_request=update_request.to_dict(), session=db_session)
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name)
        )
        for dataset_replica in db_session.execute(stmt).scalars():
            assert dataset_replica['bytes'] == len(files) * file_size
            assert dataset_replica['length'] == len(files)
            assert dataset_replica['available_bytes'] == len(files) * file_size
            assert dataset_replica['available_replicas_cnt'] == len(files)
            assert dataset_replica['state'] == ReplicaState.AVAILABLE

        # Delete first replica on first RSE -> replica on first RSE should be unavailable, replica on second RSE should be still available
        delete_replicas(rse_id=rse_id, files=[files[0]], session=db_session)
        db_session.commit()
        models.UpdatedCollectionReplica(scope=mock_scope, name=dataset_name, did_type=constants.DIDType.DATASET).save(session=db_session)
        # delete_replica creates also update object but with rse_id -> extra filter for rse_id is NULL
        stmt = select(
            models.UpdatedCollectionReplica
        ).where(
            and_(models.UpdatedCollectionReplica.scope == mock_scope,
                 models.UpdatedCollectionReplica.name == dataset_name,
                 models.UpdatedCollectionReplica.rse_id.is_(None))
        )
        update_request = db_session.execute(stmt).scalar_one()
        update_collection_replica(update_request=update_request.to_dict(), session=db_session)
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name,
                 models.CollectionReplica.rse_id == rse_id)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == len(files) * file_size
        assert dataset_replica['length'] == len(files)
        assert dataset_replica['available_bytes'] == (len(files) - 1) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files) - 1
        assert dataset_replica['state'] == ReplicaState.UNAVAILABLE
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name,
                 models.CollectionReplica.rse_id == rse2_id)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == len(files) * file_size
        assert dataset_replica['length'] == len(files)
        assert dataset_replica['available_bytes'] == len(files) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files)
        assert dataset_replica['state'] == ReplicaState.AVAILABLE

        # Set the state of the first replica on the second RSE to UNAVAILABLE -> both replicass should be unavailable
        stmt = select(
            models.RSEFileAssociation
        ).where(
            and_(models.RSEFileAssociation.rse_id == rse2_id,
                 models.RSEFileAssociation.scope == mock_scope,
                 models.RSEFileAssociation.name == files[0]['name'])
        )
        file_replica = db_session.execute(stmt).scalar_one()
        file_replica.state = constants.ReplicaState.UNAVAILABLE
        models.UpdatedCollectionReplica(scope=mock_scope, name=dataset_name, did_type=constants.DIDType.DATASET).save(session=db_session)
        stmt = select(
            models.UpdatedCollectionReplica
        ).where(
            and_(models.UpdatedCollectionReplica.scope == mock_scope,
                 models.UpdatedCollectionReplica.name == dataset_name,
                 models.UpdatedCollectionReplica.rse_id.is_(None))
        )
        update_request = db_session.execute(stmt).scalar_one()
        update_collection_replica(update_request=update_request.to_dict(), session=db_session)
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name,
                 models.CollectionReplica.rse_id == rse_id)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == len(files) * file_size
        assert dataset_replica['length'] == len(files)
        assert dataset_replica['available_bytes'] == (len(files) - 1) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files) - 1
        assert dataset_replica['state'] == ReplicaState.UNAVAILABLE
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name,
                 models.CollectionReplica.rse_id == rse2_id)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == len(files) * file_size
        assert dataset_replica['length'] == len(files)
        assert dataset_replica['available_bytes'] == (len(files) - 1) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files) - 1
        assert dataset_replica['state'] == ReplicaState.UNAVAILABLE

        # Delete first replica on second RSE -> file is not longer part of dataset -> both replicas should be available
        delete_replicas(rse_id=rse2_id, files=[files[0]], session=db_session)
        db_session.commit()
        models.UpdatedCollectionReplica(scope=mock_scope, name=dataset_name, did_type=constants.DIDType.DATASET).save(session=db_session)
        stmt = select(
            models.UpdatedCollectionReplica
        ).where(
            and_(models.UpdatedCollectionReplica.scope == mock_scope,
                 models.UpdatedCollectionReplica.name == dataset_name,
                 models.UpdatedCollectionReplica.rse_id.is_(None))
        )
        update_request = db_session.execute(stmt).scalar_one()
        update_collection_replica(update_request=update_request.to_dict(), session=db_session)
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name,
                 models.CollectionReplica.rse_id == rse_id)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == (len(files) - 1) * file_size
        assert dataset_replica['length'] == len(files) - 1
        assert dataset_replica['available_bytes'] == (len(files) - 1) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files) - 1
        assert dataset_replica['state'] == ReplicaState.AVAILABLE
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name,
                 models.CollectionReplica.rse_id == rse2_id)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == (len(files) - 1) * file_size
        assert dataset_replica['length'] == len(files) - 1
        assert dataset_replica['available_bytes'] == (len(files) - 1) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files) - 1
        assert dataset_replica['state'] == ReplicaState.AVAILABLE

        # Add first replica to the first RSE -> first replicas should be available
        add_replicas(rse_id=rse_id, files=[files[0]], account=root_account, session=db_session)
        attach_dids(scope=mock_scope, name=dataset_name, dids=[files[0]], account=root_account, session=db_session)
        models.UpdatedCollectionReplica(scope=mock_scope, name=dataset_name, did_type=constants.DIDType.DATASET).save(session=db_session)
        stmt = select(
            models.UpdatedCollectionReplica
        ).where(
            and_(models.UpdatedCollectionReplica.scope == mock_scope,
                 models.UpdatedCollectionReplica.name == dataset_name,
                 models.UpdatedCollectionReplica.rse_id.is_(None))
        )
        update_request = db_session.execute(stmt).scalar_one()
        update_collection_replica(update_request=update_request.to_dict(), session=db_session)
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name,
                 models.CollectionReplica.rse_id == rse_id)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == len(files) * file_size
        assert dataset_replica['length'] == len(files)
        assert dataset_replica['available_bytes'] == len(files) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files)
        assert dataset_replica['state'] == ReplicaState.AVAILABLE
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name,
                 models.CollectionReplica.rse_id == rse2_id)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == len(files) * file_size
        assert dataset_replica['length'] == len(files)
        assert dataset_replica['available_bytes'] == (len(files) - 1) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files) - 1
        assert dataset_replica['state'] == ReplicaState.UNAVAILABLE

        # Add first replica to the second RSE -> both replicas should be available again
        add_replicas(rse_id=rse2_id, files=[files[0]], account=root_account, session=db_session)
        models.UpdatedCollectionReplica(scope=mock_scope, name=dataset_name, did_type=constants.DIDType.DATASET).save(session=db_session)
        stmt = select(
            models.UpdatedCollectionReplica
        ).where(
            and_(models.UpdatedCollectionReplica.scope == mock_scope,
                 models.UpdatedCollectionReplica.name == dataset_name,
                 models.UpdatedCollectionReplica.rse_id.is_(None))
        )
        update_request = db_session.execute(stmt).scalar_one()
        update_collection_replica(update_request=update_request.to_dict(), session=db_session)
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name,
                 models.CollectionReplica.rse_id == rse_id)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == len(files) * file_size
        assert dataset_replica['length'] == len(files)
        assert dataset_replica['available_bytes'] == len(files) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files)
        assert dataset_replica['state'] == ReplicaState.AVAILABLE
        stmt = select(
            models.CollectionReplica
        ).where(
            and_(models.CollectionReplica.scope == mock_scope,
                 models.CollectionReplica.name == dataset_name,
                 models.CollectionReplica.rse_id == rse2_id)
        )
        dataset_replica = db_session.execute(stmt).scalar_one()
        assert dataset_replica['bytes'] == len(files) * file_size
        assert dataset_replica['length'] == len(files)
        assert dataset_replica['available_bytes'] == len(files) * file_size
        assert dataset_replica['available_replicas_cnt'] == len(files)
        assert dataset_replica['state'] == ReplicaState.AVAILABLE



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from rucio.common import exception
from rucio.common.constants import DEFAULT_VO
from rucio.core import exporter
from rucio.core.rse import get_rse_name
from rucio.db.sqla.constants import DatabaseOperationType
from rucio.db.sqla.session import db_session
from rucio.gateway import permission


def export_data(issuer: str, distance: bool = True, vo: str = DEFAULT_VO) -> dict[str, Any]:
    """
    Export data from Rucio.

    :param issuer: the issuer.
    :param distance: To enable the reporting of distance.
    :param vo: the VO of the issuer.
    """
    kwargs = {'issuer': issuer}
    with db_session(DatabaseOperationType.READ) as session:
        auth_result = permission.has_permission(issuer=issuer, vo=vo, action='export', kwargs=kwargs, session=session)
        if not auth_result.allowed:
            raise exception.AccessDenied('Account %s can not export data. %s' % (issuer, auth_result.message))

        data = exporter.export_data(distance=distance, vo=vo, session=session)
        rses = {}
        distances = {}

        for rse_id in data['rses']:
            rse = data['rses'][rse_id]
            rses[get_rse_name(rse_id=rse_id, session=session)] = rse
        data['rses'] = rses

        if distance:
            for src_id in data['distances']:
                dests = data['distances'][src_id]
                src = get_rse_name(rse_id=src_id, session=session)
                distances[src] = {}
                for dest_id in dests:
                    dest = get_rse_name(rse_id=dest_id, session=session)
                    distances[src][dest] = dests[dest_id]
            data['distances'] = distances
    return data



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from json import dumps
from typing import Optional

from requests import post

from rucio.client.baseclient import BaseClient, choice
from rucio.common.constants import DEFAULT_VO
from rucio.common.exception import RucioException, UnsupportedDIDType


class TouchClient(BaseClient):

    """
    Touch client class to send a trace that can be used to
    update accessed_at for file or dataset DIDs
    """

    DIDS_BASEURL = 'dids'
    TRACES_BASEURL = 'traces'

    def touch(
            self,
            scope: str,
            name: str,
            rse: Optional[str] = None
    ) -> None:
        """
        Sends a touch trace for a given file or dataset.

        Parameters
        ----------
        scope : The scope of the file/dataset to update.
        name : The name of file/dataset to update.
        rse : Optional parameter if a specific replica should be touched.

        Raises
        ------
        DataIdentifierNotFound
            If given DIDs does not exist.
        RSENotFound
            If rse is not None and given rse does not exist.
        UnsupportedDIDType
            If type of the given DID is not FILE or DATASET.
        RucioException
            If trace could not be sent successfully.
        """

        trace = {}

        trace['eventType'] = 'touch'
        trace['clientState'] = 'DONE'
        trace['account'] = self.account
        if self.vo != DEFAULT_VO:
            trace['vo'] = self.vo

        if rse:
            self.get_rse(rse)  # pylint: disable=no-member

            trace['localSite'] = trace['remoteSite'] = rse

        info = self.get_did(scope, name)  # pylint: disable=no-member

        if info['type'] == 'CONTAINER':
            raise UnsupportedDIDType("%s:%s is a container." % (scope, name))

        if info['type'] == 'FILE':
            trace['scope'] = scope
            trace['filename'] = name
        elif info['type'] == 'DATASET':
            trace['datasetScope'] = scope
            trace['dataset'] = name

        url = '%s/%s/' % (choice(self.list_hosts), self.TRACES_BASEURL)

        try:
            post(url, verify=False, data=dumps(trace))
        except Exception as error:
            raise RucioException("Could not send trace. " + str(error))
